/*%
 *  TITULO:	Fixed point
 *  ARCHIVO:	FixPoint.cpp
 *
 *  VERSION:	SVC 0.0.0
 *
 *  CONTENIDO:
 *
 *  AUTORES:	Daniel Bacher
 *  FECHA:	Tue  09-05-1995  db
 *  REFERENCIA:
 *
 *		Copyright (c) 1994-1995 Soluciones Verticales Consultores
 *		All rights reserved.
%*/
#include <stdio.h>
#include "fixpoint.h"
#include <assert.h>
#include <string.h>
#include <math.h>
#include <ctype.h>

const unsigned long powers [7] =
{
      1l
    , 10l
    , 100l
    , 1000l
    , 10000l
    , 100000l
    , 1000000l
};

FixPoint::FixPoint()
{
    hi = lo = 0;
    negative = 0;
    point = 2;
}

FixPoint::FixPoint (long double TheValue)
{
    point = 2;
    if (0 > TheValue)
    {
	negative = 1;
	TheValue = -TheValue;
    }
    else
    {
	negative = 0;
    }
    TheValue *= powers [point];
    hi = TheValue / LOMAX;
    lo = fmod (TheValue, LOMAX);
}

FixPoint::FixPoint (const FixPoint& TheValue)
{
    *this = TheValue;
}

/*% Sintaxis:	    FixPoint::FixPoint (const char *pszValue)
 *
 *  Descripcin:    Converts a string to a fixed point value.
 *
 *		    Parmetro	    Tipo/Descripcin
 *		    ---------	    ----------------
 *		    pszValue	    const char *, The String to be
 *				    converted
 *
 *  Retorna:
 *		    Valor	    Significado
 *		    -----	    -----------
 *
 *  Comentarios:    Using regular expresions the string should match:
 *
 *		    [-+]?[0-9]+(\.[0-9]+)?
 *
 *		    i.e. an optional '+' or '-' sign followed by one or
 *		    more digits, optionally followed by a period and one
 *		    or more digits.
 *
 *  Historia:	    Tue  09-05-1995  db
 *		    No '+' sign is handling
 *
 *  Implementacin: 1.	Find the period position within the string
 *		    2.	Add whole digits
 *		    3.	Add fractional digits
%*/
FixPoint::FixPoint (const char *pszValue)
{
    int PointOffset = 0;
    int length;
    for (length = 0; '\0' != pszValue [length]; length++)
    {
	if ('.' == pszValue [length])
	{
	    PointOffset = length;
	}
    }

    point = length - PointOffset - 1;
    unsigned long base = 1;
    lo = 0;
    hi = 0;
    int i;
    for (i = length; 0 < i; i--)
    {
	if (isdigit (pszValue [i - 1]))
	{
	    if (base < LOMAX)
	    {
		lo += base * (pszValue [i - 1] - '0');
		base *= 10;
	    }
	    else
	    {
		break;
	    }
	}
    }
    base = 1;
    for ( ; 0 < i; i--)
    {
	if (isdigit (pszValue [i - 1]))
	{
	    if ((base < HIMAX) || ('0' != pszValue [i - 1]))
	    {
		hi += base * (pszValue [i - 1] - '0');
		base *= 10;
	    }
	    else
	    {
		// error overflow!
		assert ((base < HIMAX) || ('0' != pszValue [i - 1]));
		break;
	    }
	}
    }
    negative = ('-' == pszValue [0]) ? 1 : 0;
}

FixPoint::~FixPoint ()
{
}

FixPoint &FixPoint::operator = (const FixPoint &TheValue)
{
    hi = TheValue.hi;
    lo = TheValue.lo;
    negative = TheValue.negative;
    point = TheValue.point;
    return *this;
}

/*% Sintaxis:	    FixPoint &FixPoint::operator += (const FixPoint &TheValue)
 *
 *  Descripcin:
 *
 *		    Parmetro	    Tipo/Descripcin
 *		    ---------	    ----------------
 *
 *  Retorna:
 *		    Valor	    Significado
 *		    -----	    -----------
 *
 *  Comentarios:    If the signs are equal do an addition.
 *		    if the signs are different substract the smaller
 *		    value from the larger one. Keep the sign from the
 *		    larger value.
 *
 *		    Check for:
 *		    -	Overflow
 *		    -	Sign after the operation
 *
 *  Historia:	    Tue  09-05-1995  db.
 *
 *  Implementacin:
%*/
FixPoint &FixPoint::operator += (const FixPoint &TheValue)
{
    assert (point == TheValue.point);
    if (TheValue.negative == negative)
    {
	Add (TheValue);
    }
    else if (TheValue.IsGreaterThan (*this))
    {
	FixPoint Temp (TheValue);
	Temp.Sub (*this);
	*this = Temp;
    }
    else
    {
	Sub (TheValue);
    }
    return *this;
}

/*% Sintaxis:	    FixPoint &FixPoint::operator -= (const FixPoint &TheValue)
 *
 *  Descripcin:
 *
 *		    Parmetro	    Tipo/Descripcin
 *		    ---------	    ----------------
 *
 *  Retorna:
 *		    Valor	    Significado
 *		    -----	    -----------
 *
 *  Comentarios:
 *
 *  Historia:	    Tue  09-05-1995  db
 *
 *  Implementacin: 1.	If the signs are different
 *		    1.1.    Add to the current value the new value.
 *		    2.	else
 *		    2.1.    If the other value is greater
 *		    2.1.1.	The result is the copy of the other
 *				value minus this one.
 *		    2.2.    else
 *		    2.2.1.	Substract from this value the other one.
%*/
FixPoint &FixPoint::operator -= (const FixPoint &TheValue)
{
    assert (point == TheValue.point);
    if (TheValue.negative != negative)
    {
	Add (TheValue);
    }
    else if (TheValue.IsGreaterThan (*this))
    {
	FixPoint Temp (TheValue);
	Temp.Sub (*this);
	*this = -Temp;
    }
    else
    {
	Sub (TheValue);
    }
    return *this;
}

/*% Sintaxis:	    void FixPoint::Add (const FixPoint &TheValue)
 *
 *  Descripcin:    Adds two values, ignoring signs.
 *
 *		    Parmetro	    Tipo/Descripcin
 *		    ---------	    ----------------
 *
 *  Retorna:
 *		    Valor	    Significado
 *		    -----	    -----------
 *
 *  Comentarios:
 *
 *  Historia:	    Wed  09-06-1995  db
 *
 *  Implementacin:
%*/
void FixPoint::Add (const FixPoint &TheValue)
{
    if (LOMAX - lo < TheValue.lo)
    {
	if (1 <= HIMAX - hi)
	{
	    hi++;
	    lo += TheValue.lo - LOMAX;
	}
	else 
	{
	    // error overflow!
	    assert (1 <= HIMAX - hi);
	}
    }
    else
    {
	lo += TheValue.lo;
    }
    if (TheValue.hi <= HIMAX - hi)
    {
	hi += TheValue.hi;
    }
    else
    {
	// error overflow!
	assert (TheValue.hi <= HIMAX - hi);
    }
}

/*% Sintaxis:	    void FixPoint::Sub (const FixPoint &TheValue)
 *
 *  Descripcin:    Decrements this value by TheValue, ignoring signs.
 *
 *		    Parmetro	    Tipo/Descripcin
 *		    ---------	    ----------------
 *
 *  Retorna:
 *		    Valor	    Significado
 *		    -----	    -----------
 *
 *  Comentarios:    The following must be true: TheValue <= *this.
 *
 *  Historia:	    Wed  09-06-1995  db
 *
 *  Implementacin:
%*/
void FixPoint::Sub (const FixPoint &TheValue)
{
    assert (!IsLessThan (TheValue));
    if (TheValue.lo > lo)
    {
	hi--;
	lo = lo - TheValue.lo + LOMAX;
    }
    else
    {
	lo -= TheValue.lo;
    }
    hi -= TheValue.hi;
}

/*% Sintaxis:	    int  FixPoint::operator > (const FixPoint &TheValue) const
 *
 *  Descripcin:
 *
 *		    Parmetro	    Tipo/Descripcin
 *		    ---------	    ----------------
 *
 *  Retorna:
 *		    Valor	    Significado
 *		    -----	    -----------
 *
 *  Comentarios:    'a' is bigger than 'b' if:
 *		    -	'a' is poisitve y 'b' is negative.
 *		    -	'a' and 'b' are negative and 'a' is closer to 0.
 *		    -	'a' and 'b' are positive and 'b' is closer to 0.
 *
 *  Historia:
 *
 *  Implementacin:
%*/
int  FixPoint::operator > (const FixPoint &TheValue) const
{
    int Result;

    assert (TheValue.point == point);
    if (negative && TheValue.negative)
    {
	Result = TheValue.IsGreaterThan (*this); 
    }
    else if (IsPositive () && TheValue.IsPositive ())
    {
	Result = IsGreaterThan (TheValue);
    }
    else 
    {
	Result = (IsPositive () && TheValue.negative);
    }
    return Result;
}

int  FixPoint::operator <  (const FixPoint &TheValue) const
{
    int Result;

    assert (TheValue.point == point);
    if (negative && TheValue.negative)
    {
	Result = TheValue.IsLessThan (*this);
    }
    else if (IsPositive () && TheValue.IsPositive ())
    {
	Result = IsLessThan (TheValue);
    }
    else 
    {
	Result = (negative && TheValue.IsPositive ());
    }
    return Result;
}

/*% Sintaxis:	    int FixPoint::IsGreaterThan (const FixPoint &TheValue) const
 *
 *  Descripcin:    Checks to see if the other value is closer to 0 than
 *		    this value, ignoring the signs.
 *
 *		    TheValue.
 *
 *		    Parmetro	    Tipo/Descripcin
 *		    ---------	    ----------------
 *
 *  Retorna:
 *		    Valor	    Significado
 *		    -----	    -----------
 *
 *  Comentarios:
 *
 *  Historia:	    Wed  09-06-1995  db
 *
 *  Implementacin:
%*/
int FixPoint::IsGreaterThan (const FixPoint &TheValue) const
{
    return ((hi > TheValue.hi) || ((hi == TheValue.hi) && (lo > TheValue.lo)));
}

/*% Sintaxis:	    int FixPoint::IsLessThan (const FixPoint &TheValue) const
 *
 *  Descripcin:    Checks wheter the current value is closer to 0 than
 *		    the other value.
 *
 *		    Parmetro	    Tipo/Descripcin
 *		    ---------	    ----------------
 *
 *  Retorna:
 *		    Valor	    Significado
 *		    -----	    -----------
 *
 *  Comentarios:    Signs are ignored.
 *
 *  Historia:
 *
 *  Implementacin:
%*/
int FixPoint::IsLessThan (const FixPoint &TheValue) const
{
    return ((hi < TheValue.hi) || ((hi == TheValue.hi) && (lo < TheValue.lo)));
}

int  FixPoint::operator == (const FixPoint &TheValue) const
{
    FixPoint Result = *this - TheValue;
    return Result.IsZero ();
}

const	FixPoint &operator + (const FixPoint &Left, const FixPoint &Right)
{
    static FixPoint Result;
    Result = Left;
    Result += Right;
    return Result;
}

const	FixPoint &operator - (const FixPoint &Left, const FixPoint &Right)
{
    static FixPoint Result;
    Result = Left;
    Result -= Right;
    return Result;
}

/*% Sintaxis:	    FixPoint &FixPoint::operator *= (const FixPoint &TheValue)
 *
 *  Descripcin:
 *
 *		    Parmetro	    Tipo/Descripcin
 *		    ---------	    ----------------
 *
 *  Retorna:
 *		    Valor	    Significado
 *		    -----	    -----------
 *
 *  Comentarios:    Note that the product is like integer multiplication
 *		    but the point would be moved to:
 *
 *		    point + TheValue.point.
 *
 *  Historia:	    Wed  09-06-1995  db.
 *
 *  Implementacin: 1.	Multiply the members based on integer
 *			multiplication.
 *		    2.	Move the result theValue.point positions.
 *		    3.	Adjust signs.
%*/
FixPoint &FixPoint::operator *= (const FixPoint &TheValue)
{
    FixPoint Result;

    Result.hi = hi * TheValue.hi * LOMAX;
    Result.hi = hi * TheValue.lo;
    Result.hi = lo * TheValue.ho;
    Result.lo = lo * TheValue.lo;
    Result.lo /= powers [TheValue.point];
    Result.lo += (Result.hi % powers [TheValue.point]) * (LOMAX / powers [TheValue.point]);
    Result.hi /= powers [TheValue.point];

    *this = Result;
    return *this;
}

const	FixPoint &operator * (const FixPoint &Left, const FixPoint &Right)
{
    static FixPoint Result;
    Result = Left;
    Result *= Right;
    return Result;
}

/*% Sintaxis:	    FixPoint &FixPoint::operator /= (const FixPoint &TheValue)
 *
 *  Descripcin:    Not really implemented.
 *
 *		    Parmetro	    Tipo/Descripcin
 *		    ---------	    ----------------
 *
 *  Retorna:
 *		    Valor	    Significado
 *		    -----	    -----------
 *
 *  Comentarios:    I will do this someday...
 *
 *  Historia:	    Mon  09-11-1995  db
 *
 *  Implementacin: 
%*/
FixPoint &FixPoint::operator /= (const FixPoint &TheValue)
{
#if 0
    assert (TheValue.point == point);
    const long double Dividend = hi * LOMAX + lo;
    const long double Divider = TheValue.hi * LOMAX + TheValue.lo;
    const long double Result = Dividend / Divider;

    *this = FixPoint (Result);
#endif
    return *this;
}

const	FixPoint &FixPoint::operator / (const FixPoint &Left, const FixPoint &Right)
{
    static FixPoint Result;
    Result = Left;
    Result /= Right;
    return Result;
}

/*% Sintaxis:	    void FixPoint::ToString (char *pszBuffer, size_t SizeofBuffer) const
 *
 *  Descripcin:    Converts a fixed point value to a string.
 *
 *		    Parmetro	    Tipo/Descripcin
 *		    ---------	    ----------------
 *
 *  Retorna:
 *		    Valor	    Significado
 *		    -----	    -----------
 *
 *  Comentarios:    Note that the original sprintf format does not work
 *		    with CodeWarrior DR6 (Bug report has been filed).
 *
 *  Historia:	    Wed  09-06-1995  db
 *
 *  Implementacin:
%*/
void FixPoint::ToString (char *pszBuffer, size_t SizeofBuffer) const
{
    assert (NULL != pszBuffer);
    pszBuffer [SizeofBuffer - 1] = '\0';

    unsigned long low = lo / powers [point];
    unsigned long fraction = lo % powers [point];
    char pszSign [2] = {'+', '\0'};
    pszSign [0] = (negative) ? '-' : '\0';
    if (0 != hi)
    {
#if 0
	sprintf (pszBuffer, "%s%lu%-0lu%c%02lu", pszSign, hi, low, '.', fraction);
#else
	sprintf (pszBuffer, "%s%lu%-0lu", pszSign, hi, low);
	sprintf (pszBuffer + strlen (pszBuffer), "%c%02lu", '.', fraction);
#endif	
    }
    else
    {
#if 0    
	sprintf (pszBuffer, "%s%lu%c%02lu", pszSign, low, '.', fraction);
#else
	sprintf (pszBuffer, "%s%lu", pszSign, low);
	sprintf (pszBuffer + strlen (pszBuffer), "%c%02lu", '.', fraction);
#endif	
    }
    assert ('\0' == pszBuffer [SizeofBuffer - 1]);
}

FixPoint::FixPoint (const TwoLongs& TheValue)
{
    lo = TheValue.lo & 0x3fffffffl;
    hi = TheValue.hi & 0x3fffffffl;
    negative = (0 != (TheValue.hi & 0x80000000l));
    point = ((TheValue.hi >> 30) << 2) + (TheValue.lo >> 30);
}

const TwoLongs &FixPoint::ToTwoLongs ()
{
    static TwoLongs TheValue;

    TheValue.hi = hi | (((negative << 1) | (point >> 2)) << 30);
    TheValue.lo = hi | ((point & 0x3) << 30);
    return TheValue;
}

const FixPoint &FixPoint::Percent (unsigned int TheValue) const
{
    static FixPoint Result;
    Result.hi *= TheValue;
    Result.lo *= TheValue;

    Result.lo /= 100;
    Result.lo += (Result.hi % 100) * (LOMAX / 100);
    Result.hi /= 100;

    return Result;
}
